# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from libqtile.scripts.migrations._base import (
    Check,
    RenamerTransformer,
    _QtileMigrator,
    add_migration,
)


class RenameTileMasterTransformer(RenamerTransformer):
    from_to = ("masterWindows", "master_length")


class RenameTileMaster(_QtileMigrator):
    ID = "RenameTileMaster"

    SUMMARY = "Changes ``masterWindows`` argument to ``master_length``."

    HELP = """
    To be consistent with other layouts, the ``masterWindows`` property of the
    ``Tile`` layout was renamed to ``master_length``. Configs using the
    ``masterWindows`` argument when configuring the layout should replace this.
    """

    AFTER_VERSION = "0.16.1"

    TESTS = [
        Check(
            """
            from libqtile import layout
            from libqtile.layout import Tile

            layouts = [
                Tile(masterWindows=2),
                layout.Tile(masterWindows=3)
            ]
            """,
            """
            from libqtile import layout
            from libqtile.layout import Tile

            layouts = [
                Tile(master_length=2),
                layout.Tile(master_length=3)
            ]
            """,
        )
    ]

    visitor = RenameTileMasterTransformer


add_migration(RenameTileMaster)
